# -*- coding: utf-8 -*-
from .finder import FinderModule

class SecurityModule(object):
    '''
    This class facilitates methods for securing the AWS account

    Methods are available to help improve AWS account security by detecting bad configurations.
    '''

    @classmethod
    def get_instance_portlisting(cls, aws, instanceid):
        '''
        List SecurityGroups and rules for an instance

        Args:
            aws: AwsManager client
            instanceid: Id of instance to analyze

        Return:
            Dictionary with instance and its SecurityGroups
        '''
        results = dict()
        instance = FinderModule(aws).find_instance(aws, 'id', instanceid)
        results.update(instance)

        results['SecurityGroups'] = list()
        for secgroup in instance["SecurityGroups"]:
            sg = secgroup
            sg['Rules'] = list()
            for rule in aws.ec2.get_secgroup_by('id', secgroup["GroupId"])["IpPermissions"]:
                sg['Rules'].append({
                    'FromPort' : rule.get("FromPort", ''),
                    'ToPort'   : rule.get("ToPort", ''),
                    'Protocol' : rule.get("IpProtocol", '') if rule.get("IpProtocol", '') != '-1' else 'ALL',
                    'IpRange'  : [iprange["CidrIp"] for iprange in rule.get("IpRanges", '')]
                })
            results['SecurityGroups'].append(sg)

        return {'Instance': results}

    @classmethod
    def get_region_portlisting(cls, aws, region):
        '''
        List SecurityGroups and rules for all instances in region

        Args:
            aws: AwsManager client
            region: Region to analyze

        Return:
            Dictionary with regions, instances and its SecurityGroups
        '''
        results = []
        aws.ec2.change_region(region)

        for instance in aws.ec2.get_instances():
            ins_element = dict()
            ins_element.update(instance)
            ins_element['SecurityGroups'] = list()
            for securitygroup in instance['SecurityGroups']:
                sg_element = dict()
                sg_element.update(securitygroup)
                sg_element['Rules'] = list()
                for rule in aws.ec2.get_secgroup_by('id', securitygroup["GroupId"])["IpPermissions"]:
                    sg_element['Rules'].append({'ToPort'   : rule.get("ToPort", ''),
                                                'FromPort' : rule.get("FromPort", ''),
                                                'Protocol' : rule.get("IpProtocol", '') if rule.get("IpProtocol", '') != '-1' else 'ALL',
                                                'IpRange'  : [iprange["CidrIp"] for iprange in rule.get("IpRanges", '')]})
                ins_element['SecurityGroups'].append(sg_element)
            results.append(ins_element)

        return {'RegionName': region, 'Instances': results}



# -*- coding: utf-8 -*-
from .base import AwsBase
from threading import Lock

class RdsService(AwsBase):
    '''
    Class belonging to the Remote Database System service.
    '''

    database_filters = {
        'id': 'db-instance-id',
        'cluster': 'db-cluster-id',
    }

    def _extract_databases(self, filters=[], regions=[], return_first=False):
        results = dict() if return_first else list()
        regions = self.parse_regions(regions=regions)
        lock = Lock()
        
        def worker(region):
            lock.acquire()
            self.change_region(region['RegionName'])
            config = self.get_client_vars()
            lock.release()
            
            rdss = self.client.describe_db_instances(Filters=filters)['DBInstances']

            if rdss:
                rds = self.inject_client_vars(rdss, config)
                if return_first:
                    results.update(rdss[0])
                else:
                    results.extend(rdss)

        for region in regions:
            self.pool.add_task(worker, region)
        self.pool.wait_completion()

        return results


    def get_database_by(self, filters, regions=[]):
        formatted_filters = self.validate_filters(filters, self.database_filters)
        return self._extract_databases(filters=formatted_filters, regions=regions, return_first=True)



    def get_databases(self, regions=[]):
        '''
        Get RDS instances in regions

        Args:
            regions (list): Regions where you want to look for

        Returns:
            (list): List of RDS dicts
        '''
        return self._extract_databases(regions=regions)

    def get_snapshots(self, regions=[]):
        '''
        Get RDS snapshots in regions

        Args:
            regions (list): Regions where you want to look for

        Returns:
            (list): List of RDS dicts
        '''
        results = list()
        regions = self.parse_regions(regions=regions)
        lock = Lock()
        
        def worker(region):
            lock.acquire()
            self.change_region(region['RegionName'])
            config = self.get_client_vars()
            lock.release()
            rdss = self.client.describe_db_snapshots()['DBSnapshots']
            results.extend(self.inject_client_vars(rdss, config))

        for region in regions:
            self.pool.add_task(worker, region)
        self.pool.wait_completion()

        return results


    def __init__(self):
        AwsBase.__init__(self, 'rds')



# -*- coding: utf-8 -*-
from .base import AwsBase
from threading import Lock
import dns.resolver


class ElbService(AwsBase):
    '''
    Class belonging to the Load Balancers service.
    '''

    loadbalancer_filters = {
        'domain': '',
        'tagname': '',
        'cname': '',
    }

    @classmethod
    def _get_cname_from_domain(cls, domain):
        '''
        Get CNAME from a domain

        Raises:
            dns.resolver.NXDOMAIN: DNS Name not registered.
            dns.resolver.NoAnswer: DNS Name not found.

        Return:
            str: String with DNS Canonical Name.
        '''
        try:
            cname = str(dns.resolver.query(domain, "CNAME")[0]).rstrip('.')
            if 'aws.com' not in cname: raise ValueError('Domain %s is not in AWS' % domain)
            return cname
        except (dns.resolver.NXDOMAIN):
            raise dns.resolver.NXDOMAIN("Couldn't find any records(NXDOMAIN)")
        except (dns.resolver.NoAnswer):
            raise dns.resolver.NoAnswer("Couldn't find any records (NoAnswer)")


    def get_loadbalancers(self, regions=[]):
        '''
        Get all Elastic Load Balancers for a region

        Args:
            regions (list): Regions where to look for this element

        Returns:
            LoadBalancers (list): List of dictionaries with the load balancers requested
        '''
        results = list()
        regions = self.parse_regions(regions=regions)
        lock = Lock()
        
        def worker(region):
            lock.acquire()
            self.change_region(region['RegionName'])
            config = self.get_client_vars()
            lock.release()
            elbs = self.client.describe_load_balancers()['LoadBalancerDescriptions']
            results.extend(self.inject_client_vars(elbs, config))

        
        [self.pool.add_task(worker, region) for region in regions]
        self.pool.wait_completion()

        return results

    def get_loadbalancers_by(self, filter_key, filter_value, regions=[]):
        '''Get loadbalancers which match with the filters
        
        Args:
            filter_key (str): [description]
            filter_value (str): [description]
            regions (list, optional): Defaults to []. List of regions to search in
        
        Returns:
            list: List of load balancers requested
        '''

        results = list()
        
        # TODO - Raruno (by Oscar)
        if filter_key == 'tagname':
            for region in self.parse_regions(regions=regions):
                try:
                    self.change_region(region)
                    results.extend(self.client.describe_load_balancers(LoadBalancerNames=[filter_value])['LoadBalancerDescriptions'])
                except:
                    pass

        return results


    def get_loadbalancer_by(self, filter_key, filter_value, regions=[]):
        '''
        Get a load balancer for a region that matches with filter

        Args:
            filter_key (str): Name of the filter
            filter_value (str): Value of the filter
            regions (list): Regions where to look for this element

        Raises:
            dns.resolver.NXDOMAIN: DNS Name not registered.
            dns.resolver.NoAnswer: DNS Name not found.

        Return:
            LoadBalancer (dict): Dictionary with the load balancer requested
        '''
        if filter_key not in self.loadbalancer_filters:
            raise ValueError('Invalid filter key. Allowed filters: ' + str(self.loadbalancer_filters.keys()))

        #bucle infinito
        if filter_key == 'tagname':
            elbs = self.get_loadbalancer_by(filter_key, filter_value, regions)
        else:
            cname = self._get_cname_from_domain(filter_value) if filter_key == 'domain' else filter_value
            self.change_region(cname.split('.')[1])
            elbs = [elb for elb in self.get_loadbalancers() if elb['DNSName'].lower() == cname.lower()]

        return next(iter(elbs or []), None)


    def __init__(self):
        AwsBase.__init__(self, 'elb')




address_filters = {
    'publicip': 'public-ip',
    'privateip': 'private-ip-address',
    'domain': 'domain',
    'instance': 'instance-id'
}


def _extract_addresses(self, filters=[], regions=[], return_first=False):
    results = list()

    regions = self.parse_regions(regions)
    for region in regions:
        self.change_region(region['RegionName'])

        addresses = self.client.describe_addresses(Filters=filters)['Addresses']
        addresses = self.inject_client_vars(addresses)

        if return_first and addresses:
            return addresses[0]

        results.extend(addresses)

    return results

def get_addresses(self, regions=[]):
    '''
    Get all IP Addresses for a region

    Args:
        regions (lst): Regions where to look for this element

    Returns:
        Addresses (dict): List of dictionaries with the addresses requested
    '''
    return self._extract_addresses(regions=regions)

def get_addresses_by(self, filters, regions=[]):
    '''
    Get all IP Addresses for a region

    Args:
        regions (lst): Regions where to look for this element

    Returns:
        Addresses (dict): List of dictionaries with the addresses requested
    '''
    formatted_filters = self.validate_filters(filters, self.address_filters)
    return self._extract_addresses(filters=formatted_filters, regions=regions)

def get_address_by(self, filters, regions=[]):
    '''
    Get IP Addresses for a region that matches with filters

    Args:
        regions (lst): Regions where to look for this element

    Returns:
        Address (dict): Dictionary with the address requested
    '''
    formatted_filters = self.validate_filters(filters, self.address_filters)
    return self._extract_addresses(filters=formatted_filters, regions=regions, return_first=True)




# -*- coding: utf-8 -*-
from .services import *

class ServiceManager:
    '''
    Parent class that provides access to services.

    For each service (ec2, s3, vpc ...) you are given access through a property of this class.
    This property will return an instance of the corresponding class, for example Ec2Service or VpcService.
    Each class of service (Ec2Service, S3Service ...) inherits from the AwsBase class.
    '''

    _ec2 = None
    _elb = None
    _acm = None
    _iam = None
    _rds = None
    _s3 = None
    _ce = None
    _route53 = None

    @property
    def ec2(self):
        if self._ec2 is None: self._ec2 = Ec2Service()
        return self._ec2

    @property
    def elb(self):
        if self._elb is None: self._elb = ElbService()
        return self._elb

    @property
    def acm(self):
        if self._acm is None: self._acm = AcmService()
        return self._acm

    @property
    def iam(self):
        if self._iam is None: self._iam = IamService()
        return self._iam

    @property
    def rds(self):
        if self._rds is None: self._rds = RdsService()
        return self._rds

    @property
    def s3(self):
        if self._s3 is None: self._s3 = S3Service()
        return self._s3

    @property
    def ce(self):
        if self._ce is None: self._ce = CostExplorerService()
        return self._ce

    @property
    def route53(self):
        if self._route53 is None: self._route53 = Route53Service()
        return self._route53


    @classmethod
    def get_auth_config(cls):
        '''
        Get the configuration of the client currently configured

        This method allows us to work with multiple accounts and different authentication methods (keys and profiles) without getting lost.

        Returns:
            A dictionary with the type of authentication used and the associated value. The secret_key is not returned for security reasons.

            {'Authorization': {'Type': 'Profile', 'Value': 'MyBoringCompany'}}
        '''
        if AwsBase.profile:
            auth = {'Authorization': {'Type':'Profile', 'Value':AwsBase.profile}}
        elif AwsBase.access_key:
            auth = {'Authorization': {'Type':'AccessKeys', 'Value':AwsBase.access_key}}
        else:
            auth = {'Authorization': {'Type':'Profile', 'Value': 'default'}}
        return auth


    def __init__(self, region, profile=None, access_key=None, secret_key=None):
        '''
        Constructor of the parent class of the services.

        With this method you can modify the configuration of the awspice client.
        It allows us to change the profile, the region or the access codes.

        Args:
            region (str): Region in which to make queries and operations.
            profile (str): Name of the AWS profile set in ~/.aws/credentials file
            access_key (str): API access key of your AWS account
            secret_key (str): API secret key of your AWS account
        '''
        AwsBase.region = region
        AwsBase.access_key = access_key
        AwsBase.secret_key = secret_key
        AwsBase.profile = profile




def _extract_vpcs(self, filters=[], regions=[], return_first=False):
    results = list()

    regions = self.parse_regions(regions)
    for region in regions:
        self.change_region(region['RegionName'])

        vpcs = self.client.describe_vpcs(Filters=filters)['Vpcs']
        vpcs = self.inject_client_vars(vpcs)

        if return_first and vpcs:
            self.change_region(curRegion)
            return vpcs[0]

        results.extend(vpcs)

    return results

def get_vpcs(self, regions=[]):
    '''
    Get all VPCs for a region

    Returns:
        VPCs (lst): List of dictionaries with the vpcs requested
    '''
    return self._extract_vpcs(regions=regions)

def get_default_vpc(self):
    '''
    Get default Security Group

    Returns:
        SecurityGroup (dict): Default security group resource
    '''
    vpcs = self.get_vpcs()

    vpc = next((x for x in vpcs if x['IsDefault'] is True), None)
    
    return vpc



# -*- coding: utf-8 -*-
from .base import AwsBase
from datetime import datetime

class IamService(AwsBase):
    '''
    Class belonging to the IAM Identity & Access management service.
    '''

    def get_inactive_users(self):
        ''' Get users who have not logged in AWS since 1 year.
        This method returns users who haven't used their password and one of their keys in less than 9 months.
        
        Returns:
            list: List of inactive users
        '''

        results = []
        today = datetime.now()
        min_inactive_days = 270 # 9 months

        def worker(user):
            pass_last_use = user.get('PasswordLastUsed', None)
            pass_inactive_days = (today - pass_last_use.replace(tzinfo=None)).days if pass_last_use else None

            if not pass_last_use or pass_inactive_days > min_inactive_days:

                inactive_keys = []
                for key in self.client.list_access_keys(UserName=user['UserName'])['AccessKeyMetadata']:
                    key_info = self.client.get_access_key_last_used(AccessKeyId=key['AccessKeyId'])['AccessKeyLastUsed']
                    
                    key_last_use = key_info.get('LastUsedDate', None)
                    key_inactive_days = (today - key_last_use.replace(tzinfo=None)).days if key_last_use else None

                    if not key_last_use or key_inactive_days > min_inactive_days:
                        key.update({'Inactive': not bool(key_last_use), 'LastUsed': key_last_use})
                        inactive_keys.append(key)
                
                if inactive_keys:
                    inactive_pass = {'Inactive': not bool(pass_last_use), 'LastUsed': pass_last_use}
                    inactive_user = { 'LoginActivity' : { 'Password' : inactive_pass, 'AccessKeys' : inactive_keys } }
                    user.update(inactive_user)
                    results.append(user)


        for user in self.get_users(): self.pool.add_task(worker, user)
        self.pool.wait_completion()
        
        return results

    def get_users(self):
        '''
        List all users for an AWS account

        Returns:
            List of all users
        '''
        results = []
        config = self.get_client_vars()

        paginator = self.client.get_paginator('list_users').paginate()
        for response in paginator:
            for user in response['Users']: results.append(user)

        return self.inject_client_vars(results, config)

    def get_access_keys(self, user):
        results = []
        access_keys = self.client.list_access_keys(UserName=user)['AccessKeyMetadata']

        def worker(ak):
            ak['LastUse'] = self.get_access_key_last_used(ak['AccessKeyId'])['AccessKeyLastUsed']
            results.append(ak)

        [self.pool.add_task(worker, ak) for ak in access_keys]
        self.pool.wait_completion()

        return self.inject_client_vars(results)

    def get_access_key_last_used(self, accesskey):
        last_use = self.client.get_access_key_last_used(AccessKeyId=accesskey)
        if last_use:
            last_use = self.inject_client_vars([last_use])[0]
        return last_use

    def __init__(self):
        AwsBase.__init__(self, 'iam')



# -*- coding: utf-8 -*-
from threading import Thread
import sys
if sys.version_info[:3] < (3,0):
    from Queue import Queue
else:
    from queue import Queue
from time import mktime
import json
import datetime
import socket

class Worker(Thread):
    """
    Thread executing tasks from a given tasks queue
    http://code.activestate.com/recipes/577187-python-thread-pool/
    """
    def __init__(self, tasks):
        Thread.__init__(self)
        self.tasks = tasks
        self.daemon = True
        self.start()

    def run(self):
        while True:
            func, args, kargs = self.tasks.get()
            try: func(*args, **kargs)
            except Exception as e: 
                print(e)
            self.tasks.task_done()

class ThreadPool:
    """
    Pool of threads consuming tasks from a queue
    http://code.activestate.com/recipes/577187-python-thread-pool/
    """
    def __init__(self, num_threads):
        self.tasks = Queue(num_threads)
        for _ in range(num_threads): Worker(self.tasks)

    def add_task(self, func, *args, **kargs):
        """Add a task to the queue"""
        self.tasks.put((func, args, kargs))

    def wait_completion(self):
        """Wait for completion of all the tasks in the queue"""
        self.tasks.join()


class ClsEncoder(json.JSONEncoder):
    '''
    JSON encoder extension.

    Sometimes Boto3 returns a non-serializable result to JSON and we get the following error when dumping that result:
    `TypeError: datetime.datetime (2015, 12, 3, 21, 20, 17, 326000, tzinfo = tzutc ()) is not JSON serializable`
    Solve it using this class encoder in ``cls`` argument

    Examples:
        json.dumps(results, indent=4, cls=awspice.ClsEncoder)
    '''

    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return int(mktime(obj.timetuple()))

        return json.JSONEncoder.default(self, obj)


def ip_in_aws(ip):
    '''
    Check if an IP address is from AWS

    Arguments:
        ip: Address to check

    Returns:
        bool
    '''
    return bool(dnsinfo_from_ip(ip))


def extract_region_from_ip(ip):
    '''Get the region where a IP is located and if it's on AWS

    Args:
        ip (str): IP address 
    
    Returns:
        (bool, str): It's in AWS / Region where the IP is located
    '''
    hostname = dnsinfo_from_ip(ip)

    if hostname:
        region = hostname['region']
        return (True, region) if region else (True, '')
    else:
        return (False, None)



def dnsinfo_from_ip(ip):
    '''
    Returns the DNS name of an IP address

    Arguments:
        ip: Address of the element.

    Examples:
        dns = get_dnsname_from_ip('8.8.8.8')

    Returns:
        dict: {'region': 'eu-west-1', 'service': 'ec2'}
    '''
    result = {}
    service_matchs = {'compute': 'ec2'}
    try:
        host = socket.gethostbyaddr(ip)[0]
        if host.endswith('aws.com'):
            # ELASTIC IPs > name.service.aws.com (4)
            # AUTOASSIGNED IPs > name.region.service.aws.com (5)
            service = host.split('.')[2]
            if service in service_matchs.keys():
                service = service_matchs[service]

            if len(host.split('.')) == 5:
                region = host.split('.')[1]
            else:
                region = None
            
            result = {'region': region, 'service': service}
        return result
    except socket.herror:
        pass
    except socket.gaierror:
        pass
    return result



# -*- coding: utf-8 -*-
from .servicemanager import ServiceManager
from .modules import *
from botocore.exceptions import ProfileNotFound, ClientError


class AwsManager:
    """
    Main class that provides access to services (ec2, s3, vpc ...) and modules (finder, stats ..)

    This master class provides access to individual services through the "services" property,
    and also to other complex modules such as "finder", "stats" and "security".

    Attributes:
        aws: Object of type #ServiceManager that provides access to the other services.
    """

    @property
    def service(self):
        return self.aws

    @property
    def finder(self):
        return FinderModule(self.aws)

    @property
    def security(self):
        return SecurityModule(self.aws)

    @property
    def stats(self):
        return StatsModule(self.aws)


    def test(self):
        """
        Method to verify that the loaded configuration is correct and access with the AWS API is correct

        Returns:
            boolean. True if the test was successful, false if it failed.
        """
        try:
            self.aws.ec2.get_regions()
            print('[OK] Your awspice is ready to give a helping hand :)')
            return True
        except ProfileNotFound:
            print('[!] You type like a lame pigeon. Profile not found in ~/.aws/credentials file')
            return False
        except ClientError:
            print('[!] Was it so difficult to copy and paste? Invalid credentials. Verify your access keys or permissions for that IAM user')
            return False


    def __init__(self, region='eu-west-1', profile=None, access_key=None, secret_key=None):
        '''
        Initialization and configuration of the client

        Args:
            region (str): Region in which to make queries and operations.
            profile (str): Name of the AWS profile set in ~/.aws/credentials file
            access_key (str): API access key of your AWS account
            secret_key (str): API secret key of your AWS account

        Returns:
            None

        '''
        self.aws = ServiceManager(region, profile=profile, access_key=access_key, secret_key=secret_key)




snapshot_filters = {
    'id': 'snapshot-id',
    'status': 'status',
    'owner': 'owner-id',
    'volume': 'volume-id',
}


def get_snapshots(self):
    '''
    Get all snapshots owned by self for the current region

    Returns:
        Snapshots (lst): List of dictionaries with the snapshots requested
    '''
    snapshots = self.client.describe_snapshots(OwnerIds=['self'])['Snapshots']
    return self.inject_client_vars(snapshots)

def get_snapshot_by(self, filters):
    '''
    Get a snapshot for a region tha matches with filters

    Args:
        filter_key (str): Name of the filter
        filter_value (str): Value of the filter

    Returns:
        Snapshot (dict): Dictionary with the snapshot requested
    '''
    snapshots = self.get_snapshots_by(filters)
    if snapshots and snapshots:
        return self.inject_client_vars(snapshots)[0]
    return None

def get_snapshots_by(self, filters):
    '''
    Get all snapshots for the current region that matches with filters

    Args:
        filter_key (str): Name of the filter
        filter_value (str): Value of the filter

    Returns:
        Snapshots (lst): List of dictionaries with the snapshots requested
    '''
    formatted_filters = self.validate_filters(filters, self.snapshot_filters)
    return self.inject_client_vars(self.client.describe_snapshots(Filters=formatted_filters)['Snapshots'])





# -*- coding: utf-8 -*-
from awspice.services.base import AwsBase
from awspice.helpers import dnsinfo_from_ip
from botocore.exceptions import ClientError
import time

class Ec2Service(AwsBase):
    '''
    Class belonging to the EC2 Computing service.
    '''


    # #################################
    # -----------  COMMON  -----------
    # #################################

    def set_tag(self, resource_id, tag_key, tag_value, regions=[]):
        '''
        Set tag for an instance

        Args:
            elements_id (str): Id of resources to tag. (i.e: i-01234, vol-01234)
            tag_key (str): Name of the element TAG (i.e: Name)
            tag_value (str): Value of that Tag
            regions (lst): Regions where to look for this element

        Returns:
            None
        '''
        tags = [{'Key': tag_key, 'Value': tag_value}]
        curRegion = AwsBase.region
        regions = self.parse_regions(regions)

        for region in regions:
            self.change_region(region['RegionName'])
            try:
                self.client.create_tags(Resources=resource_id, Tags=tags)
                return None
            except ClientError:
                pass
            finally:
                self.change_region(curRegion)

        raise Exception('The element {} does not exist for that regions.'.format(resource_id))

    # #################################
    # -------------- AMIS -------------
    # #################################
    from ._ec2.image import ami_filters
    from ._ec2.image import ami_distributions
    from ._ec2.image import distrib_amis

    from ._ec2.image import _extract_amis
    from ._ec2.image import get_amis
    from ._ec2.image import get_ami_by
    from ._ec2.image import get_amis_by
    from ._ec2.image import get_amis_by_distribution


    # #################################
    # ----------- INSTANCES -----------
    # #################################
    from ._ec2.instance import instance_filters
    from ._ec2.instance import instance_status_filters
    
    from ._ec2.instance import _extract_instances
    from ._ec2.instance import get_instances
    from ._ec2.instance import get_instance_by
    from ._ec2.instance import get_instances_by
    from ._ec2.instance import start_instances
    from ._ec2.instance import stop_instances
    from ._ec2.instance import create_instances

    from ._ec2.instance import _extract_instance_status
    from ._ec2.instance import get_instances_status
    from ._ec2.instance import get_instance_status_by
    from ._ec2.instance import get_instances_status_by
    


    # #################################
    # ------------ VOLUMES ------------
    # #################################
    from ._ec2.volume import volume_filters

    from ._ec2.volume import _extract_volumes
    from ._ec2.volume import get_volumes
    from ._ec2.volume import get_volume_by
    from ._ec2.volume import get_volumes_by


    # #################################
    # ----------- SNAPSHOTS -----------
    # #################################
    from ._ec2.snapshot import snapshot_filters

    from ._ec2.snapshot import get_snapshots
    from ._ec2.snapshot import get_snapshot_by
    from ._ec2.snapshot import get_snapshots_by


    # #################################
    # ---------- SEC. GROUPS ----------
    # #################################
    from ._ec2.security_group import secgroup_filters

    from ._ec2.security_group import _extract_secgroups
    from ._ec2.security_group import get_secgroups
    from ._ec2.security_group import get_secgroup_by
    from ._ec2.security_group import get_secgroups_by
    from ._ec2.security_group import create_security_group
    from ._ec2.security_group import delete_security_group


    # #################################
    # ----------- ADDRESSES -----------
    # #################################
    from ._ec2.address import address_filters

    from ._ec2.address import _extract_addresses
    from ._ec2.address import get_addresses
    from ._ec2.address import get_addresses_by
    from ._ec2.address import get_address_by
    

    # #################################
    # ------------- VPCS --------------
    # #################################
    from ._ec2.vpc import _extract_vpcs
    from ._ec2.vpc import get_vpcs
    from ._ec2.vpc import get_default_vpc


    def __init__(self):
        AwsBase.__init__(self, 'ec2')



from threading import Lock

volume_filters = {
    'id': 'volume-id',
    'status': 'status',
    'instance': 'attachment.instance-id',
    'autodelete': 'attachment.delete-on-termination',
    'encrypted': 'encrypted',
    'tagname': 'tag:Name',
}


def _extract_volumes(self, filters=[], regions=[], return_first=False):
    regions = self.parse_regions(regions)
    results = dict() if return_first else list()
    lock = Lock()
    
    def worker(region):
        lock.acquire()
        self.change_region(region['RegionName'])
        config = self.get_client_vars()
        lock.release()

        volumes = self.client.describe_volumes(Filters=filters)['Volumes']

        if volumes:
            volumes = self.inject_client_vars(volumes, config)
            if return_first:
                results.update(volumes[0])
            else:
                results.extend(volumes)

    # Launch tasks in threads
    for region in regions: self.pool.add_task(worker, region=region)

    # Wait results
    self.pool.wait_completion()

    return results

def get_volumes(self, regions=[]):
    '''
    Get all volumes for one or more regions

    Args:
        regions (lst): Regions where to look for this element

    Returns:
        Volumes (lst): List of dictionaries with the volumes requested
    '''
    return self._extract_volumes(regions=regions)

def get_volume_by(self, filters, regions=[]):
    '''
    Get a volume for one or more regions that matches with filters

    Args:
        filter_key (str): Name of the filter
        filter_value (str): Value of the filter
        regions (lst): Regions where to look for this element

    Returns:
        Volume (dict): Dictionary with the volume requested
    '''
    return self.get_volumes_by(filters, regions, return_first=True)

def get_volumes_by(self, filters, regions=[], return_first=False):
    '''
    Get volumes for one or more regions that matches with filters

    Args:
        filter_key (str): Name of the filter
        filter_value (str): Value of the filter
        regions (lst): Regions where to look for this element

    Returns:
        Volume (dict): Dictionary with the volume requested
    '''
    formatted_filters = self.validate_filters(filters, self.volume_filters)
    return self._extract_volumes(filters=formatted_filters, regions=regions, return_first=return_first)




import sys, json
import awspice
import botocore
if sys.version_info >= (3, 3):
    from unittest.mock import patch
else:
    from mock import patch

calls = [
    'DescribeInstances',
]

def load(filepath):
    with open(filepath) as file:
        content = json.loads(file.read())
    return content

@patch("botocore.client.BaseClient._make_api_call")
def test_get_instances(mock):
    mock.return_value=load('test/data/ec2/instances.json')
    region = 'eu-west-1'
    x = awspice.connect(region).service.ec2.get_instances()
    assert all(instance['Region']['RegionName'] == region for instance in x)



from threading import Lock

secgroup_filters = {
    'id': 'group-id',
    'name': 'group-name',
    'description': 'description',
    'protocol': 'ip-permission.protocol',
    'fromport': 'ip-permission.from-port',
    'toport': 'ip-permission.to-port',
    'range': 'ip-permission.cidr',
}

def _extract_secgroups(self, filters=[], regions=[], return_first=False):
    regions = self.parse_regions(regions)
    results = dict() if return_first else list()
    lock = Lock()

    def worker(region):
        lock.acquire()
        self.change_region(region['RegionName'])
        config = self.get_client_vars()
        lock.release()

        secgroups = self.client.describe_security_groups(Filters=filters)["SecurityGroups"]

        if secgroups:
            secgroups = self.inject_client_vars(secgroups, config)

            if return_first:
                results.update(secgroups[0])
            if not return_first:
                results.extend(secgroups)

    for region in regions: self.pool.add_task(worker, region=region)
    self.pool.wait_completion()
    
    return results

def get_secgroups(self, regions=[]):
    '''
    Get all security groups for the current region

    Returns:
        SecurityGroups (lst): List of dictionaries with the security groups requested
    '''
    return self._extract_secgroups(regions=regions)

def get_secgroup_by(self, filters, regions=[]):
    '''
    Get security group for a region that matches with filters

    Args:
        filter_key (str): Name of the filter
        filter_value (str): Value of the filter

    Returns:
        SecurityGroup (dict): Dictionaries with the security group requested
    '''
    formatted_filters = self.validate_filters(filters, self.secgroup_filters)
    return self._extract_secgroups(filters=formatted_filters, regions=regions, return_first=True)

def get_secgroups_by(self, filters, regions=[]):
    '''
    Get all security groups for a region that matches with filters

    Args:
        filter_key (str): Name of the filter
        filter_value (str): Value of the filter

    Returns:
        SecurityGroups (lst): List of dictionaries with the security groups requested
    '''
    formatted_filters = self.validate_filters(filters, self.secgroup_filters)
    return self._extract_secgroups(filters=formatted_filters, regions=regions)

def create_security_group(self, name, allowed_range, vpc_id=None):
    '''
    Create a new Security Group

    Args:
        name (str): Name of the Security Group
        allowed_range (str): Network range with permissions (i.e: 10.0.0.0/32)
        vpc_id (str): Id of assigned VPC

    Returns:
        str: Identifier of the security group created.
    '''
    vpc = vpc_id if vpc_id else self.get_default_vpc()['VpcId']
    desc = 'Security group created by Awspice'
    default_rules = [
        {'IpProtocol': 'tcp',
            'FromPort': 80,
            'ToPort': 80,
            'IpRanges': [{'CidrIp': allowed_range}]},
        {'IpProtocol': 'tcp',
            'FromPort': 22,
            'ToPort': 22,
            'IpRanges': [{'CidrIp': allowed_range}]}
    ]

    sg_id = self.client.create_security_group(
        GroupName=name, VpcId=vpc, Description=desc)['GroupId']
    self.client.authorize_security_group_ingress(
        GroupId=sg_id, IpPermissions=default_rules)

    return sg_id

def delete_security_group(self, identifier):
    '''
    Delete an existing Security Group

    Args:
        identifier (str): Id of the Security Group

    Returns:
        none
    '''
    self.client.delete_security_group(GroupId=identifier)



ami_filters = {
    'id': 'image-id',
    'name': 'name',
    'architecture': 'architecture',
    'platform': 'platform',
    'owner': 'owner-id',
    'public': 'is-public',
    'state': 'state',
}

ami_distributions = {
    'ubuntu': 'ubuntu/images/hvm-ssd/ubuntu-*-*{version}*-amd64-server-*',
    'windows': 'Windows_Server-*{version}*-English-*-Base-20*.*.*',
    'amazon': 'amzn-ami-hvm-20*.*.*-x86_64-*',
}

distrib_amis = {
    'ubuntu': 'ami-f90a4880',
    'windows': 'ami-b5530b5e',
    'redhat': 'ami-c86c3f23',
}


def _extract_amis(self, filters=[], regions=[], return_first=False):
    filters.append({'Name': 'state', 'Values': ['available', 'pending']})
    # Just supported x64 OS
    filters.append({'Name': 'architecture', 'Values': ['x86_64']})
    filters.append({'Name': 'hypervisor', 'Values': ['xen']})
    filters.append({'Name': 'virtualization-type', 'Values': ['hvm']})
    filters.append({'Name': 'image-type', 'Values': ['machine']})
    filters.append({'Name': 'root-device-type', 'Values': ['ebs']})

    curRegion = self.region
    regions = self.parse_regions(regions)
    results = list()

    for region in regions:
        self.change_region(region['RegionName'])

        amis = self.client.describe_images(Filters=filters)['Images']
        amis = self.inject_client_vars(amis)
        if return_first and amis:
            self.change_region(curRegion)
            return amis[0]
        results.extend(amis)

    self.change_region(curRegion)
    return results

def get_amis_by_distribution(self, distrib, version='*', latest=False, regions=[]):
    '''
    Get one or more Images filtering by distribution

    Args:
        distrib (str): Distribution of the image (i.e.: ubuntu)
        version (str): Version of the system
        latest (bool): True if only returns the newest item.
        regions (lst): Regions where to look for this element

    Return:
        Image (lst): List with the images requested.
    '''

    self.validate_filters(distrib, self.ami_distributions.keys())
    filters = [
        {'Name': 'name', 'Values': [self.ami_distributions[distrib].format(version=version)]},
        {'Name': 'is-public', 'Values': ['true']}
    ]

    results = self._extract_amis(filters=filters, regions=regions)
    results = sorted(results, key=lambda k: k['Name'])

    if latest and results:
        return [results[-1]]

    return results

def get_ami_by(self, filters, regions=[]):
    '''
    Get an ami for one or more regions that matches with filter

    Args:
        filter_key (str): Name of the filter
        filter_value (str): Value of the filter
        regions (lst): Regions where to look for this element

    Return:
        Image (dict): Image requested
    '''
    return self.get_amis_by(filters=filters,
                            regions=regions,
                            return_first=True)

def get_amis_by(self, filters, regions=[], return_first=False):
    '''
    Get list of amis for one or more regions that matches with filter

    Args:
        filter_key (str): Name of the filter
        filter_value (str): Value of the filter
        regions (lst): Regions where to look for this element
        return_first (bool): True if return first result

    Return:
        Images (lst): List of requested images
    '''
    formatted_filters = self.validate_filters(filters, self.ami_filters)
    return self._extract_amis(filters=formatted_filters, regions=regions, return_first=return_first)
    
def get_amis(self, regions=[]):
    '''
    Get all images

    Args:
        regions (lst): Regions where to look for this element

    Returns:
        Images (lst): List of all images
    '''
    return self._extract_amis(regions=regions)


# -*- coding: utf-8 -*-

class StatsModule:
    '''
    Class responsible for processing general data to the AWS account.

    This class is dedicated to the global management of the AWS account in order to obtain statistics, costs or global information.

    Attributes:
        aws: awspice client
    '''

    def get_stats(self, regions=[]):
        '''
        Retrieve data about services in your AWS account like Volumes, Instances or Databases.

        Args:
            aws: AwsManager client
            region: To retrieve data only of this region

        Return:
            List of regions with its stats
        '''
        results = dict()
        regions = [region.encode('ascii', 'ignore') for region in regions]

        results['Users'] = self.aws.iam.get_users()
        results['Buckets'] = self.aws.s3.get_buckets()
        results['Regions'] = dict()
        for region in regions:
            self.aws.ec2.change_region(region)
            data = dict()
            data['Instances'] = self.aws.ec2.get_instances()
            data['SecurityGroups'] = self.aws.ec2.get_secgroups()
            data['Volumes'] = self.aws.ec2.get_volumes()
            # data['Snapshots'] = self.aws.ec2.get_snapshots();  # Need to select only private snaps
            data['Addresses'] = self.aws.ec2.get_addresses()
            data['Vpcs'] = self.aws.ec2.get_vpcs()
            data['LoadBalancers'] = self.aws.elb.get_loadbalancers()
            data['Databases'] = self.aws.rds.get_databases()
            data['Certificates'] = self.aws.acm.list_certificates()
            results['Regions'][region] = data
        return results

    def cost_saving(self, regions=[]):
        '''
        List unused elements that carry expenses.

        Args:
            aws: AwsManager client.

        Returns:
            Dict Region with a list of regions with its unused elements
        '''
        results = dict()
        regions = [region.encode('ascii', 'ignore') for region in regions]

        for region in regions:
            self.aws.ec2.change_region(region)

            savings = dict()
            savings['Volumes'] = self.aws.ec2.get_volumes_by({'status': 'available'})
            adds = self.aws.ec2.get_addresses()
            savings['Addresses'] = filter(lambda x: x.get('AssociationId') == None, adds)
            elbs = self.aws.elb.get_loadbalancers()
            savings['LoadBalancers'] = filter(lambda x: x.get('Instances') == [], elbs)

            results[region] = savings
        return {'Regions': results}

    def __init__(self, aws):
        self.aws = aws



# -*- coding: utf-8 -*-
import json
import boto3
from awspice.helpers import ThreadPool
from pkg_resources import resource_filename

class AwsBase(object):
    '''
    Base class from which all services inherit (ec2, s3, vpc ...)

    This class contains methods and properties that are common to all AWS services and should
    be accessible by all of them. This class is responsible for instantiating the client and
    processing information related to the accounts and regions.

    Attributes:
        client: Boto3 client
        region: Current region used by the client
        profile: Current profile used by the client
        access_key: Current access key used by the client
        secret_key: Current secret key used by the client

    '''
    endpoints = None
    region = None
    profile = None
    access_key = None
    secret_key = None

    # THREADS NUMBER
    pool = ThreadPool(30)

    service_resources = ['ec2', 's3']


    def set_client(self, service):
        '''
        Main method to set Boto3 client

        Args:
            service (str): Service to use    (i.e.: ec2, s3, vpc...)
            region (str): Region name to use (i.e.: eu-central-1)
            profile (str): Profile name set in ~/.aws/credentials file
            access_key (str): API access key of your AWS account
            secret_key (str): API secret key of your AWS account

        Raises:
            ClientError: Access keys are not valid or lack of permissions for a service/region
            ProfileNotFound: Profile name not found in credentials file

        Returns:
            None
        '''
        _region = str(AwsBase.region)
        _profile = str(AwsBase.profile) if AwsBase.profile else None
        _access_key = str(AwsBase.access_key) if AwsBase.access_key else None
        _secret_key = str(AwsBase.secret_key) if AwsBase.secret_key else None

        # 1. Validate args for authentication
        self.set_auth_config(region=_region,
                             profile=_profile,
                             access_key=_access_key,
                             secret_key=_secret_key)
        # 2. Set Boto3 client
        if _profile:
            self.client = boto3.Session(profile_name=_profile).client(service, region_name=_region)
            if service in self.service_resources:
                self.resource = boto3.Session(profile_name=_profile).resource(
                    service, region_name=_region)

        elif _access_key and _secret_key:
            self.client = boto3.client(service,
                                    region_name=_region,
                                    aws_access_key_id=_access_key,
                                    aws_secret_access_key=_secret_key)
            if service in self.service_resources:
                self.resource = boto3.resource(service,
                                            region_name=_region,
                                            aws_access_key_id=_access_key,
                                            aws_secret_access_key=_secret_key)
        # If auth isn't provided, set "default" profile (.aws/credentials)
        else:
            self.client = boto3.client(service, region_name=_region)
            if service in self.service_resources:
                self.resource = boto3.resource(service, region_name=_region)

    @classmethod
    def set_auth_config(cls, region, profile=None, access_key=None, secret_key=None):
        '''
        Set properties like service, region or auth method to be used by boto3 client

        Args:
            service (str): Service to use               (i.e.: ec2, s3, vpc...)
            region (str): Region name                   (i.e.: eu-central-1)
            access_key (str): API Access key
            secret_key (str): API Secret key
            profile (str): Profile name set in ~/.aws/credentials file
        '''
        AwsBase.region = region

        if profile and (access_key or secret_key):
            raise ValueError('Use Profile or Access keys, not both.')

        if profile:
            AwsBase.profile = profile
            AwsBase.access_key = None
            AwsBase.secret_key = None
        elif access_key and secret_key:
            AwsBase.profile = None
            AwsBase.access_key = access_key
            AwsBase.secret_key = secret_key

    @classmethod
    def get_client_vars(cls):
        '''Get information of the current client configuration
        Sometimes we need to store this variables, for example using threads, 
        because AwsBase is constantly changing 
        
        Returns:
            dict: Array with current client configuration ({'region': 'eu-west-1', 'profile': 'default'})
        '''

        _region_name = str(AwsBase.region)
        _region = dict(AwsBase.endpoints['Regions'][_region_name], RegionName=_region_name)
        _profile = str(AwsBase.profile)
        _access_key = str(AwsBase.access_key)

        return {'region': _region, 'profile': _profile, 'access_key': _access_key}



    @classmethod
    def inject_client_vars(cls, elements, client_conf=None):
        '''
        Insert in each item of a list, the region and the current credentials.

        This function is called by all the methods of all the services that return a list of objects
        to identify in what region and account they have been found.

        Args:
            elements (list): List of dictionaries
            client_conf (dict): Array with the client configuration (see `get_client_vars`)

        Returns:
            list. Returns same list with the updated elements (region and authentication included)

        '''

        # [!] used dict() to avoid to rewrite object AwsBase in next line
        if client_conf:
            _region_name = client_conf['region']['RegionName']
            _region_dict = client_conf['region']
            _profile =     client_conf['profile']
            _access_key =  client_conf['access_key']
        else:
            _region_name = str(AwsBase.region)
            _region_dict = dict(AwsBase.endpoints['Regions'][_region_name])
            _profile =     str(AwsBase.profile)
            _access_key =  str(AwsBase.access_key)
            _region_dict['RegionName'] = _region_name
        results = []

        for element in elements:

            if element.get('Authorization') and element.get('RegionName'):
                break
                
            elements_tagname = filter(lambda x: x['Key'] == 'Name', element.get('Tags', ''))
            element['TagName'] = next(iter(map(lambda x: x.get('Value', ''), elements_tagname)), '')
            element['Region'] = _region_dict

            if _profile:
                element['Authorization'] = {'Type':'Profile', 'Value': _profile}
            elif _access_key:
                element['Authorization'] = {'Type':'AccessKeys', 'Value': _access_key}
            else:
                element['Authorization'] = {'Type':'Profile', 'Value': 'default'}

            results.append(element)

        return results

    def region_in_regions(self, region, regions):
        '''
        Check if region is in a complex list of regions

        Args:
            region (str | lst) - Region name or parsed region format {'RegionName': 'eu-west-1'}
            regions (lst) - List of strings or dicts of regions

        Examples:
            region_in_regions('eu-west-1', [{'RegionName': 'eu-west-1}])

        Returns:
            bool
        '''
        return self.parse_regions(region)[0] in self.parse_regions(regions)

    @classmethod
    def validate_filters(cls, input_filters, accepted_filters):
        '''
        Transform filters into AWS filters format after validate them.

        Args:
            input_filters (str): Items to validate
            accepted_filters (list): Pre-validated list

        Returns:
            None

        Raises:
            ValueError: Filter is not in the accepted filter list
        '''
        formatted_filters = {}

        for key, value in input_filters.items():
            if key in accepted_filters:
                if isinstance(value, list):
                    formatted_filters.update({'Name': accepted_filters[key], 'Values': value})
                else:
                    formatted_filters.update({'Name': accepted_filters[key], 'Values': [value]})
            else:
                raise ValueError('Invalid filter key. Allowed filters: ' + str(accepted_filters.keys()))
        
        return [formatted_filters]

            


    # #################################
    # ------------ PROFILES -----------
    # #################################

    @classmethod
    def get_profiles(cls):
        '''
        Get a list of all available profiles in ~/.aws/credentials file

        Returns:
            list. List of strings with available profiles
        '''
        return boto3.Session().available_profiles

    def change_profile(self, profile):
        '''
        Change profile of the client

        This method changes the account/profile used but keeps the same region and service

        Args:
            profile (str): Name of the profile set in ~/.aws/credentials file

        Examples:
            $ aws = awspice.connect()
            $ aws.service.ec2.change_profile('my_boring_company')

        Returns:
            None
        '''
        if profile != AwsBase.profile:
            AwsBase.profile = profile
            self.set_client(self.service)

    def parse_profiles(self, profiles=[]):
        '''
        Validation method which get a profile or profile list and return the expected list of them

        The purpose of this method is that a user can pass different types of data as a "profile"
        argument and obtain a valid output for any method that works with this type of data.

        Args:
            profiles (list | str): String or list of string to parse

        Examples:
            $ account_str = aws.service.ec2.parse_profiles('my_company')
            $ account_lst = aws.service.ec2.parse_profiles(['my_company'])
            $ accounts_lst = aws.service.ec2.parse_profiles(['my_company', 'other_company'])

        Returns:
            list. List of a strings with profile names
        '''
        if isinstance(profiles, str):
            if profiles == 'ALL':
                return self.get_profiles()
            return [profiles]

        if isinstance(profiles, list) and profiles:
            return profiles
        return [self.profile]


    # #################################
    # ------------ REGIONS ------------
    # #################################

    @classmethod
    def _load_endpoints(cls):
        '''
        Get AWS-Standard partition of endpoints.json file (botocore)

        Returns:
            dict: AWS-Standard partition
        '''
        if AwsBase.endpoints == None:
            # Load endpoints file
            endpoint_resource = resource_filename('botocore', 'data/endpoints.json')
            with open(endpoint_resource, 'r') as f:
                data = f.read()
                endpoints = json.loads(data)

            # Get regions for "AWS Standard" (Not Gov, China)
            partition = next((x for x in endpoints['partitions'] if x['partitionName'] == 'AWS Standard'), None)

            # Format JSON & Save
            results = dict()
            results['Regions'] = dict()
            results['Services'] = partition['services']
            results['Defaults'] = partition['defaults']
            results['DnsSuffix'] = partition['dnsSuffix']
            for k, v in partition['regions'].items():
                desc = v['description']
                results['Regions'][k] = {"Description": desc,
                                         "Country": desc[desc.find("(")+1:desc.find(")")]}
            AwsBase.endpoints = results

        return AwsBase.endpoints


    def get_endpoints(self):
        '''
        Get services and its regions and endpoints

        Returns:
            dict: Dict with services (key) and its regions and Endpoints.
        '''
        partition = self._load_endpoints()
        dns_suffix = partition['DnsSuffix']
        default_url = partition['Defaults']['hostname']

        # Return as list
        results = dict()
        for k, v in partition['Services'].items():
            url = v.get('Defaults', {}).get('hostname', default_url)
            results[k] = {
                'Regions': v['endpoints'].keys(),
                'Endpoints': [url.format(service=k, region=region, dnsSuffix=dns_suffix) for region in v['endpoints'].keys()],
            }
        return results

    def get_regions(self):
        '''
        Get all available regions

        Returns:
            list. List of regions with 'Country' and 'RegionName'
        '''
        return self._load_endpoints()['Regions']

    def change_region(self, region):
        '''
        Change region of the client

        This method changes the region used but keeps the same service and profile

        Args:
            region (str): Region Name (ID) of AWS (i.e.: eu-central-1)

        Examples:
            aws.service.ec2.change_region('eu-west-1')

        Returns:
            None
        '''
        AwsBase.region = region
        #raruno
        self.set_client(self.service)

    def parse_regions(self, regions=[], default_all=False):
        '''
        Validation method which get a region or list of regions and return the expected list of them

        The purpose of this method is that a user can pass different types of data as a "region"
        argument and obtain a valid output for any method that works with this type of data.

        Args:
            regions (list | str): String or list of string to parse
            default_all (bool): If the list of regions is empty and this argument is True,
                                a list with all regions will be returned. This is useful when you do
                                not know the data entry of type "region" and you want to search by
                                default in all regions (if regions are empty means that the user
                                does not know where an element is located).

        Examples:
            AwsBase.region = aws.service.ec2.parse_regions([])
            regions = aws.service.ec2.parse_regions('eu-west-1')
            regions = aws.service.ec2.parse_regions(['eu-west-1'])
            regions = aws.service.ec2.parse_regions(['eu-west-1', 'eu-west-2'])

        Returns:
            list. List of a strings with profile names
        '''
        results = list()

        # regions = 'eu-west-1'
        if isinstance(regions, str):
            results = [{'RegionName':regions}]

        # regions = ['eu-west-1'] or [u'eu-west-1'] or [{'RegionName': 'eu-west-1}]
        elif isinstance(regions, list) and regions:
            if isinstance(regions[0], dict) and regions[0].get('RegionName', False):
                return regions
            elif isinstance(regions[0], str) or isinstance(regions[0], unicode):
                [results.append({'RegionName': region}) for region in set(regions)]
            else:
                raise ValueError('Invalid regions value.')

        # regions = {'eu-west-1': {...}, 'eu-central-1' : {...} }  <--- AwsBase.endpoints['Regions']
        elif isinstance(regions, dict) and regions:
            if regions.get('eu-west-1', False):
                [results.append({'RegionName': region}) for region in regions.keys()]
            else:
                raise ValueError('Invalid regions value.')

        # regions = None <--- Get current or all regions
        else:
            results = self.get_regions() if default_all else [{'RegionName': AwsBase.region}]

        return results


    def __init__(self, service):
        '''
        This constructor configures the corresponding service according to the class that calls it.

        Every time the EC2Service Class is called (inherits from this class), this constructor will
        change the client's service to 'ec2'. And then, if ELBService service is called, this method
        is called again changing the service from 'ec2' to 'elb'.

        Args:
            service (str): AWS service to uso

        Returns:
            None
        '''
        self.service = service
        self.set_client(service=service)
        self._load_endpoints()




from threading import Lock
from awspice.helpers import extract_region_from_ip

instance_filters = {
    'id': 'instance-id',
    'dnsname': 'dns-name',
    'publicip': 'network-interface.association.public-ip',
    'privateip': 'private-ip-address',
    'name': 'tag:Name',
    'tagname': 'tag:Name',
    'status': 'instance-state-name',
    'user': 'key-name',
}
instance_status_filters = {
    'event': 'event.code',
    'status': 'instance-state-name',
    'instance-check': 'instance-status.status',
    'system-check': 'system-status.status',
}


def _extract_instances(self, filters=[], regions=[], return_first=False):
    regions = self.parse_regions(regions)
    results = dict() if return_first else list()
    lock = Lock()

    def worker(region):
        # Race Condition: Locking AwsBase.region...
        # Change region > Get client config > Do the query
        lock.acquire()
        self.change_region(region['RegionName'])
        config = self.get_client_vars()
        lock.release()

        reservations = self.client.describe_instances(Filters=filters)["Reservations"]
        for reserv in reservations:
            instances = self.inject_client_vars(reserv['Instances'], config)

            if return_first and instances:
                results.update(instances[0])
            if not return_first and instances:
                results.extend(instances)

    # Launch tasks in threads
    for region in regions: 
        self.pool.add_task(worker, region=region)

    # Wait results
    self.pool.wait_completion()
    
    return results

def get_instances(self, regions=[]):
    '''
    Get all instances for one or more regions.

    Args:
        regions (lst): Regions where to look for this element

    Returns:
        Instances (lst): List of dictionaries with the instances requested
    '''
    return self._extract_instances(regions=regions)

def get_instance_by(self, filters, regions=[]):
    '''
    Get an instance for one or more regions that matches with filter

    Args:
        filter_key (str): Name of the filter
        filter_value (str): Value of the filter
        regions (lst): Regions where to look for this element

    Return:
        Instance (dict): Dictionary with the instance requested
    '''
    return self.get_instances_by(filters, regions, return_first=True)

def get_instances_by(self, filters, regions=[], return_first=False):
    '''
    Get an instance for one or more regions that matches with filter

    Args:
        filter_key (str): Name of the filter
        filter_value (str): Value of the filter
        regions (lst): Regions where to look for this element
        return_first (bool): Select to return the first match

    Return:
        Instances (lst): List of dictionaries with the instances requested
    '''
    formatted_filters = self.validate_filters(filters, self.instance_filters)
    
    if "publicip" in filters.keys():
        ip_in_aws, ip_region = extract_region_from_ip(filters['publicip'])
        
        if not ip_in_aws:
            return {}

        # ['eu-west-1', 'eu-west-2']    / 'eu-west-1'
        # [ {'RegionName': 'eu-west-1'}, {...} ]
        if isinstance(regions, list):
            if ip_region in regions or ip_region in regions[0].values():
                region = ip_region
                
        # {'eu-west-1': {...}, 'eu-west-2': {...} }
        if isinstance(regions, dict):
            if ip_region in regions.keys():
                region = ip_region

    return self._extract_instances(filters=formatted_filters, regions=regions, return_first=return_first)

def create_instances(self, name, key_name, allowed_range, ami=None, distribution=None,
                        version=None, instance_type='t2.micro', region=None, vpc=None, count=1):
    '''
    Create a new instance

    Args:
        name (str): TagName of the instance
        key_name (str): The name of the key pair (i.e: it_user)
        allowed_range (str): Network range with access to instance (i.e: 10.0.0.0/32)
        ami (str): Id of the ami (i.e: ami-12345)
        instance_type (str): Type of hardware of the instance (i.e: t2.medium)
        distribution (str): Instead of ami, select an OS: (i.e: ubuntu)
        region (str): Name of the region where  instance will be displayed
        vpc (str): VPC identifier where the instance will be deployed.
        count (int): Number of instances to launch

    Returns:
        Instances (lst): List of launched instances
    '''

    if region:
        curRegion = AwsBase.region
        self.change_region(region)

    if not vpc:
        vpc = self.get_default_vpc()['VpcId']

    if not ami:
        latest_ami = []
        if distribution and version:
            latest_ami = self.get_amis_by_distribution(distribution, version, latest=True)
        elif distribution and not version:
            latest_ami = self.get_amis_by_distribution(distribution, latest=True)

        if latest_ami:
            ami = latest_ami[0]['ImageId']
        else:
            raise ValueError("Insert a valid AMI or distribution.\n" +
                                "Parameters: Distribution={distrib}; Version={version}; ami={ami}".format(
                    distrib=distribution,
                    version=version,
                    ami=ami))

    secgroup_id = str()
    try:
        secgroup_id = self.create_security_group(name, allowed_range, vpc)
        instance = self.resource.create_instances(
            ImageId=ami,
            InstanceType=instance_type,
            KeyName=key_name,
            SecurityGroupIds=[secgroup_id],
            MaxCount=count,
            MinCount=count,
            TagSpecifications=[
                {'ResourceType': 'instance', 'Tags': [
                    {'Key': 'Name', 'Value': name}]}
            ]
        )
        return instance
    except Exception:
        if secgroup_id:
            self.delete_security_group(secgroup_id)
        raise
    finally:
        self.change_region(curRegion)


def start_instances(self, instance_ids, regions=[]):
    '''
    Stops an Amazon EC2 instance

    Args:
        instance_ids (lst): List of identifiers of instances to be started.

    Examples:
        $ aws.service.ec2.start_instances(instances=['i-001'])
        $ aws.service.ec2.start_instances(instances=['i-001', 'i-033'], regions=['eu-west-1', 'eu-central-1'])

    Returns:
        lst: List of instances to be started, with their previous and current status.
    '''
    regions = self.parse_regions(regions)
    started_instances = list()

    for region in regions:
        self.change_region(region['RegionName'])

        for instance in instance_ids:
            try:
                x = self.client.start_instances(InstanceIds=[instance])
                started_instances.extend(x['StartingInstances'])
                instance_ids.remove(instance)

            except ClientError:
                pass

    return started_instances


def stop_instances(self, instance_ids, regions=[], force=False):
    '''
    Stops an Amazon EC2 instance

    Args:
        instance_ids (lst): List of identifiers of instances to be stopped.

    Examples:
        $ aws.service.ec2.stop_instances(instances=['i-001'])
        $ aws.service.ec2.stop_instances(instances=['i-001', 'i-033'], regions=['eu-west-1', 'eu-central-1'])

    Returns:
        lst: List of instances to be stopped, with their previous and current status.
    '''
    regions = self.parse_regions(regions)
    stopped_instances = list()

    for region in regions:
        self.change_region(region['RegionName'])

        for instance in instance_ids:
            try:
                x = self.client.stop_instances(InstanceIds=[instance], Force=force)
                stopped_instances.extend(x['StoppingInstances'])
                instance_ids.remove(instance)

            except ClientError:
                pass

    return stopped_instances

# ######################## INSTANCE STATUS ########################

def _extract_instance_status(self, filters=[], regions=[], return_first=False):
    regions = self.parse_regions(regions)
    results = dict() if return_first else list()
    lock = Lock()

    def worker(region):
        lock.acquire()
        self.change_region(region['RegionName'])
        config = self.get_client_vars()
        lock.release()

        instances = self.client.describe_instance_status(Filters=filters)["InstanceStatuses"]
        if instances:
            instances = self.inject_client_vars(instances, config)

            if return_first:
                results.update(instances[0])
            if not return_first:
                results.extend(instances)

    for region in regions: self.pool.add_task(worker, region=region)
    self.pool.wait_completion()
    
    return results

def get_instances_status(self, regions=[]):
    return self._extract_instance_status(regions=regions)

def get_instance_status_by(self, filters, regions=[]):
    formatted_filters = self.validate_filters(filters, self.instance_status_filters)
    return self.get_instances_status_by(filters, regions, return_first=True)

def get_instances_status_by(self, filters, regions=[], return_first=False):
    formatted_filters = self.validate_filters(filters, self.instance_status_filters)
    return self._extract_instance_status(filters=formatted_filters, regions=regions, return_first=return_first)


from .acm import AcmService
from .base import AwsBase
from .ce import CostExplorerService
from .ec2 import Ec2Service
from .elb import ElbService
from .iam import IamService
from .rds import RdsService
from .s3 import S3Service
from .route53 import Route53Service

__all__ = ['AwsBase', 'Ec2Service', 'ElbService', 'IamService', 'RdsService', 'S3Service', 'AcmService', 'CostExplorerService', 'Route53Service']



# -*- coding: utf-8 -*-
from .base import AwsBase
import datetime

class CostExplorerService(AwsBase):
    '''
    Class belonging to the Cost Explorer service.
    '''

    granularities = ['DAILY', 'MONTHLY']
    filter_dimensions = ['AZ', 'INSTANCE_TYPE', 'LINKED_ACCOUNT', 'OPERATION', 'PURCHASE_TYPE', 'REGION', 'SERVICE', 'USAGE_TYPE', 'USAGE_TYPE_GROUP', 'RECORD_TYPE', 'OPERATING_SYSTEM', 'TENANCY', 'SCOPE', 'PLATFORM', 'SUBSCRIPTION_ID', 'LEGAL_ENTITY_NAME', 'DEPLOYMENT_OPTION', 'DATABASE_ENGINE', 'CACHE_ENGINE', 'INSTANCE_TYPE_FAMILY']
    group_dimensions = ['AZ', 'INSTANCE_TYPE', 'LEGAL_ENTITY_NAME', 'LINKED_ACCOUNT', 'OPERATION', 'PLATFORM', 'PURCHASE_TYPE', 'SERVICE', 'TAG', 'TENANCY', 'USAGE_TYPE']

    def get_cost(self, from_date=None, to_date=None, interval="Monthly", group_by='', group_by_tag_value='', filter_by={}, ec2_running_hours=False):
        '''
        Get the cost of account or its elements.

        This method obtains costs of an account/s , one or several elements (substances, balancers, addresses)
        between two dates and granularized in days or months.
        If the date is not indicated, the cost of the last month will be returned.

        Args:
            from_date (str): Date from which you want to obtain data. (Format: 2018-04-24)
            to_date (str): Date until which you want to obtain data. (Format: 2018-04-24)
            interval (str): Time interval to be analyzed. [ MONTHLY | DAILY ]
            group_by (str): Group results by ['AZ', 'INSTANCE_TYPE', 'LEGAL_ENTITY_NAME', 'LINKED_ACCOUNT', 'OPERATION', 'PLATFORM', 'PURCHASE_TYPE', 'SERVICE', 'TAG', 'TENANCY', 'USAGE_TYPE']
            group_by_tag_value (str): TAG key in case group_by set to 'TAG' (i.e. Name, Project or Environment)
            filter_by (dict): Key of the filter and value. {'TAG_NAME': ['ec2-tagname', 'LINKED_ACCOUNT: ['1234']]}

        Examples:
            get_cost(['machine-1', 'machine-2'], '2018-12-24', '2018-12-26', interval='daily')
            get_cost() # Get account cost

        Returns:
            Costs (list): List of days or months with the requested costs
        '''
        # If the dates are empty, extract data from the last month.
        if from_date is None and to_date is None:
            from_date = (datetime.datetime.now() - datetime.timedelta(1*365/12)).replace(day=1).strftime('%Y-%m-%d')
            to_date = datetime.datetime.now().replace(day=1).strftime('%Y-%m-%d')

        # Parameters validation
        try:
            datetime.datetime.strptime(from_date, '%Y-%m-%d')
            datetime.datetime.strptime(to_date, '%Y-%m-%d')
            assert interval.upper() in self.granularities, "interval. Valid values are %s" % str(self.granularities)
            assert isinstance(group_by, str), "group_by. Valid values are %s" % str(self.group_dimensions)
            assert isinstance(filter_by, dict), "filter_by. Valid values are %s" % str(self.filter_dimensions)
        except Exception as e:
            raise ValueError("Invalid parameter: " + str(e))

        timeperiod = {'Start': from_date , 'End': to_date}
        if ec2_running_hours: filter_by['USAGE_TYPE_GROUP'] = 'EC2: Running Hours'
        filters = self._format_filters(filter_by)
        groups  = self._format_groups(group_by, group_by_tag_value)

        if filters:
            results = self.client.get_cost_and_usage(TimePeriod=timeperiod,
                                                    Granularity=interval.upper(),
                                                    Filter=filters,
                                                    Metrics=['UnblendedCost'],
                                                    GroupBy=groups)
        else:
            results = self.client.get_cost_and_usage(TimePeriod=timeperiod,
                                                    Granularity=interval.upper(),
                                                    Metrics=['UnblendedCost'],
                                                    GroupBy=groups)
        return results['ResultsByTime']
        

    def _format_groups(self, group, group_by_tag_value=''):
        """Give the correct format to groups for get_cost function
        
        Args:
            group (str): Type of group to filter
        
        Returns:
            list: Well-formatted keys and values of groups 
        """
        result = []
        if group:
            if group.upper() in self.group_dimensions:
                if group.upper() == 'TAG':
                    result = [{'Type': 'TAG', 'Key': group_by_tag_value}]
                else:
                    result = [{'Type': 'DIMENSION', 'Key': group.upper()}]
            else:
                raise ValueError("%s is not a valid filter to group by" % group)
        return result


    def _format_filters(self, filters):
        """Give the correct format to filters for get_cost function
        
        Args:
            filters {dict}: Key and values of your filters

        Raises:
            ValueError: Invalid filter key passed.
        
        Returns:
            dict -- Well-formatted keys and values of filters
        """
        args = filters.keys()
        result = {}
        if len(args) == 1:
            for key, value in filters.items():
                tagvalue = value if isinstance(value, list) else [value]
                if key.upper().startswith('TAG_'):
                    tagkey = key[4:].title()
                    result['Tags'] = {'Key': tagkey, 'Values': tagvalue}
                elif key.upper() in self.filter_dimensions:
                    result['Dimensions'] = {'Key': key.upper(), 'Values': tagvalue}
                else:
                    raise ValueError('Invalid filter. Allowed filters: %s' % str(self.filter_dimensions))
        elif len(args) > 1:
            result['And'] = list()
            for key, value in filters.items():
                result['And'].append(self._format_filters({key: value}))
        return result

    def __init__(self):
        AwsBase.__init__(self, 'ce')
        self.change_region('us-east-1')



import unittest
import awspice

class ModuleFinderTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\nStarting unit tests of base service module")

    def test_get_client_vars(self):
        aws = awspice.connect('eu-west-2', profile='test')
        client_vars = aws.service.ec2.get_client_vars()
        self.assertEqual(client_vars, {'region': {'Description': 'EU (London)', 'Country': 'London', 'RegionName': 'eu-west-2'}, 'profile': 'test', 'access_key': 'None'})

    def test_set_auth_config(self):
        aws = awspice.connect('eu-west-2', profile='test')
        client_vars = aws.service.ec2.get_client_vars()
        self.assertEqual(client_vars, {'region': {'Description': 'EU (London)', 'Country': 'London', 'RegionName': 'eu-west-2'}, 'profile': 'test', 'access_key': 'None'})
        aws.service.ec2.set_auth_config(region=client_vars['region']['RegionName'], profile='test2')
        client_vars = aws.service.ec2.get_client_vars()
        self.assertEqual(client_vars, {'region': {'Description': 'EU (London)', 'Country': 'London', 'RegionName': 'eu-west-2'}, 'profile': 'test2', 'access_key': 'None'})
        aws.service.ec2.set_auth_config(region=client_vars['region']['RegionName'], access_key='key', secret_key='secret')
        client_vars = aws.service.ec2.get_client_vars()
        self.assertEqual(client_vars, {'region': {'Description': 'EU (London)', 'Country': 'London', 'RegionName': 'eu-west-2'}, 'profile': 'None', 'access_key': 'key'})

    def test_change_region(self):
        aws = awspice.connect('eu-west-2', profile='test')
        client_vars = aws.service.ec2.get_client_vars()
        self.assertEqual(client_vars, {'region': {'Description': 'EU (London)', 'Country': 'London', 'RegionName': 'eu-west-2'}, 'profile': 'test', 'access_key': 'None'})
        aws.service.ec2.change_region('eu-west-1')
        client_vars = aws.service.ec2.get_client_vars()
        self.assertEqual(client_vars, {'region': {'Description': 'EU (Ireland)', 'Country': 'Ireland', 'RegionName': 'eu-west-1'}, 'profile': 'test', 'access_key': 'None'})

    def test_change_profile(self):
        aws = awspice.connect('eu-west-2', profile='test')
        client_vars = aws.service.ec2.get_client_vars()
        self.assertEqual(client_vars, {'region': {'Description': 'EU (London)', 'Country': 'London', 'RegionName': 'eu-west-2'}, 'profile': 'test', 'access_key': 'None'})
        aws.service.ec2.change_profile('test2')
        client_vars = aws.service.ec2.get_client_vars()
        self.assertEqual(client_vars, {'region': {'Description': 'EU (London)', 'Country': 'London', 'RegionName': 'eu-west-2'}, 'profile': 'test2', 'access_key': 'None'})

    def test_get_profiles(self):
        aws = awspice.connect('eu-west-2', profile='test')
        profiles = aws.service.ec2.get_profiles()
        ss = ['test' in item or 'test2' in item for item in profiles]
        self.assertEqual(True, all(ss))

    def test_parse_profiles(self):
        aws = awspice.connect('eu-west-2', profile='test')
        profiles = aws.service.ec2.parse_profiles('ALL')
        ss = ['test' in item or 'test2' in item for item in profiles]
        self.assertEqual(True, all(ss))
        profiles = aws.service.ec2.parse_profiles(['test'])
        self.assertEqual(profiles, ['test'])
        profiles = aws.service.ec2.parse_profiles('test5')
        self.assertEqual(profiles, ['test5'])

    def test_parse_regions(self):
        aws = awspice.connect('eu-west-1', profile='test')
        regions = aws.service.ec2.parse_regions('eu-west-2')
        self.assertEqual(regions, [{'RegionName':'eu-west-2'}])
        regions = aws.service.ec2.parse_regions(['eu-west-1', 'eu-west-2'])
        ss = ['eu-west-1' in item.values() or 'eu-west-2' in item.values() for item in regions]
        self.assertEqual(True, all(ss))
        regions = aws.service.ec2.parse_regions([{'RegionName':'eu-west-1'}, {'RegionName':'eu-west-2'}])
        ss = ['eu-west-1' in item.values() or 'eu-west-2' in item.values() for item in regions]
        self.assertEqual(True, all(ss))




if __name__ == '__main__':
        unittest.main()


# -*- coding: utf-8 -*-
u"""
Copyright 2018 David Amrani Hernandez - ElevenPaths, Telefonica

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from setuptools import setup, find_packages

def read_file(filename):
    with open(filename) as f:
        return f.read()

setup(
    name='awspice',
    version=read_file('VERSION').strip(),
    install_requires=read_file('requirements.txt').splitlines(),
    packages=find_packages(),
    author='David Amrani Hernandez',
    author_email='davidmorenomad@gmail.com',
    url='https://github.com/davidmoremad',
    project_urls={
        "Documentation": "https://awspice.readthedocs.io",
        "Source Code": "https://github.com/Telefonica/awspice",
    },
    description='Awspice is a wrapper tool of Boto3 library to list inventory and manage your AWS infrastructure The objective of the wrapper is to abstract the use of AWS, being able to dig through all the data of our account. It belongs to ElevenPaths, Telefonica.',
    long_description=open('README.rst').read(),
    keywords="aws amazon cloud platform boto3 wrapper security management region account",
    license='Apache 2.0',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Testing',
    ],
)



# -*- coding: utf-8 -*-
from .base import AwsBase

class Route53Service(AwsBase):
    '''
    Class belonging to the Route 53 DNS Service
    '''

    def get_domains(self):
        '''
        Get hosted zones and its records

        Returns:
            (lst): List of Hosted Zones with Records
        '''
        domains = []
        for hz in self.list_hosted_zones():
            hz['Records'] = self.list_records(hz['Id'])
            domains.append(hz)
        return domains

    def list_hosted_zones(self):
        '''
        List all hosted zones

        Returns:
            List of hosted zones
        '''
        return self.inject_client_vars(self.client.list_hosted_zones()['HostedZones'])

    def list_records(self, hosted_zone_id):
        '''
        List all records for a hosted zone

        Args:
            hosted zone (str): The ID of the hosted zone that contains the resource record sets that you want to list

        Returns:
            List of DNS records
        '''
        records = []

        config = self.get_client_vars()
        paginator = self.client.get_paginator('list_resource_record_sets')
        for response in paginator.paginate(HostedZoneId=hosted_zone_id):
            for x in response['ResourceRecordSets']: records.append(x)

        return self.inject_client_vars(records, config)

    def list_records_by_domain(self, domain):
        '''
        List all records of a hosted-zone domain

        Args:
            domain (str): The DOMAIN name of the hosted zone that contains the resource record sets that you want to list

        Returns:
            List of DNS records
        '''
        hosted_zone = filter(lambda x: domain  + "." == x['Name'], self.list_hosted_zones())
        if hosted_zone:
            return self.list_records(hosted_zone[0]['Id'])
        return None

    def __init__(self):
        AwsBase.__init__(self, 'route53')
        self.change_region('us-east-1')



# -*- coding: utf-8 -*-
from awspice.helpers import extract_region_from_ip
from threading import Lock

class FinderModule:
    '''
    This class makes it easy to search for components in AWS.

    Attributes:
        aws: awspice client

    '''

    def find_instance(self, filters, profiles=[], regions=[]):
        '''
        Get an instance in different accounts and regions, using search filters.
        '''
        profiles = self.aws.ec2.parse_profiles(profiles)
        regions = self.aws.ec2.parse_regions(regions, True)

        # CODE SNIPPET from aws.service.ec2.instances.get_instance_by
        if "publicip" in filters.keys():
            ip_in_aws, ip_region = extract_region_from_ip(filters['publicip'])
            
            if not ip_in_aws:
                return {}

            # ['eu-west-1', 'eu-west-2']    / 'eu-west-1'
            # [ {'RegionName': 'eu-west-1'}, {...} ]
            if isinstance(regions, list):
                if ip_region in regions or ip_region in regions[0].values():
                    region = ip_region
                    
            # {'eu-west-1': {...}, 'eu-west-2': {...} }
            if isinstance(regions, dict):
                if ip_region in regions.keys():
                    region = ip_region


        for account in profiles:
            try:
                self.aws.ec2.change_profile(account)
                instance = self.aws.ec2.get_instance_by(filters, regions=regions)
                if instance: 
                    return instance
            except:
                pass
                
        return {}

    def find_instances(self, filters=None, profiles=[], regions=[]):
        '''
        Get instances in different accounts and regions, using search filters.
        '''
        results = list()
        profiles = self.aws.ec2.parse_profiles(profiles)
        regions = self.aws.ec2.parse_regions(regions, True)


        for account in profiles:
            self.aws.ec2.change_profile(account)
            if filters:
                results.extend(self.aws.ec2.get_instances_by(filters, regions=regions))
            else:
                results.extend(self.aws.ec2.get_instances(regions=regions))
        return results

    def find_volume(self, filters, profiles=[], regions=[]):
        '''
        Get a volume in different accounts and regions, using search filters.
        '''
        profiles = self.aws.ec2.parse_profiles(profiles)
        regions = self.aws.ec2.parse_regions(regions, True)

        for account in profiles:
            try:
                self.aws.ec2.change_profile(account)
                volume = self.aws.ec2.get_volume_by(filters, regions=regions)
                if volume: 
                    return volume
            except:
                pass
        return None

    def find_volumes(self, filters=None, profiles=[], regions=[]):
        '''
        Get group of volumes in different accounts and regions, using search filters.
        '''
        results = list()
        profiles = self.aws.ec2.parse_profiles(profiles)
        regions = self.aws.ec2.parse_regions(regions, True)

        for account in profiles:
            self.aws.ec2.change_profile(account)
            if filters:
                results.extend(self.aws.ec2.get_volumes_by(filters, regions=regions))
            else:
                results.extend(self.aws.ec2.get_volumes(regions=regions))
        return results


    def find_loadbalancer(self, filters, profiles=[], regions=[]):
        '''
        Get a load balancer in different accounts and regions, using search filters.
        '''
        profiles = self.aws.elb.parse_profiles(profiles)
        regions = self.aws.elb.parse_regions(regions, True)

        for account in profiles:
            try:
                self.aws.elb.change_profile(account)
                elb = self.aws.elb.get_loadbalancer_by(filters, regions=regions)
                if elb: 
                    return elb
            except:
                pass
        return None


    def find_loadbalancers(self, filter_key=None, filter_value=None, profiles=[], regions=[]):
        '''
        Get load balancers in different accounts and regions, using search filters.
        '''
        results = list()
        profiles = self.aws.elb.parse_profiles(profiles)
        regions = self.aws.elb.parse_regions(regions, True)

        for account in profiles:
            self.aws.elb.change_profile(account)
            if filter_key and filter_value:
                results.extend([self.aws.elb.get_loadbalancer_by(filter_key, filter_value, regions=regions)])
            else:
                results.extend(self.aws.elb.get_loadbalancers(regions=regions))
        return results
        
    def find_users(self, profiles=[]):
        '''
        Get IAM users in different accounts.
        '''
        results = list()
        profiles = self.aws.iam.parse_profiles(profiles)
        lock = Lock()

        def worker(profile):
            lock.acquire()
            self.aws.iam.change_profile(profile)
            lock.release()

            results.extend(self.aws.iam.get_users())

        for profile in profiles: 
            self.aws.iam.pool.add_task(worker, profile)
        self.aws.iam.pool.wait_completion()

        return results


    def find_inactive_users(self, profiles=[]):
        '''
        Get inactive users in different accounts
        '''


        results = []
        profiles = self.aws.iam.parse_profiles(profiles)

        for profile in profiles:
            self.aws.iam.change_profile(profile)
            results.extend(self.aws.iam.get_inactive_users())

        return results



    def find_buckets(self, profiles=[]):
        '''
        Search S3 buckets in different accounts.
        '''
        results = list()
        profiles = self.aws.s3.parse_profiles(profiles)
        lock = Lock()

        def worker(profile):
            lock.acquire()
            self.aws.s3.change_profile(profile)
            lock.release()
            results.extend(self.aws.s3.get_buckets())

        for profile in profiles: 
            self.aws.s3.pool.add_task(worker, profile)
        self.aws.s3.pool.wait_completion()

        return results
        
    def find_rds_databases(self, profiles=[], regions=[]):
        '''
        Get RDS databases in different accounts and regions.
        '''
        results = list()
        profiles = self.aws.ec2.parse_profiles(profiles)
        regions = self.aws.ec2.parse_regions(regions, True)

        for account in profiles:
            self.aws.rds.change_profile(account)
            results.extend(self.aws.rds.get_databases(regions))
        return results
        
    def find_rds_snapshots(self, profiles=[], regions=[]):
        '''
        Get RDS snapshots in different accounts and regions.
        '''
        results = list()
        profiles = self.aws.ec2.parse_profiles(profiles)
        regions = self.aws.ec2.parse_regions(regions, True)

        for account in profiles:
            self.aws.rds.change_profile(account)
            results.extend(self.aws.rds.get_snapshots(regions))
        return results


    def __init__(self, aws):
        self.aws = aws



# -*- coding: utf-8 -*-
from .base import AwsBase
from botocore.exceptions import ClientError
from io import StringIO

class S3Service(AwsBase):
    '''
    Class belonging to the S3 Storage service.
    '''

    def upload_string_as_file(self, bucket_name, filepath, content):
        '''
        Upload string as a file to S3 bucket

        Args:
            bucket_name (str): Name of the S3 bucket
            filepath (str): File path which will be created. (i.e. 'folder1/folder2/filename.txt')
            content (str): File content in string format.

        Returns:
            None
        '''
        data = StringIO.StringIO(content)
        self.resource.Bucket(bucket_name).put_object(Key=filepath, Body=data)


    def get_buckets(self):
        '''
        Get all buckets in S3

        Returns:
            Buckets (list): List of dictionaries with the buckets requested
        '''
        config = self.get_client_vars()
        buckets = self.client.list_buckets()['Buckets']
        return self.inject_client_vars(buckets, config)


    def get_bucket_acl(self, bucketname):
        acl = self.client.get_bucket_acl(Bucket=bucketname)['Grants']
        return self.inject_client_vars(acl)


    def get_public_buckets(self):
        '''Get all public buckets and its permissions
        
        This method returns all buckets in an AWS Account which
        have public permissions to read, write, read acl, write acl or
        even full control.

        Returns:
            Buckets-ACL (list): List of dictionaries with the buckets requested
        '''
        results = []

        # https://docs.aws.amazon.com/AmazonS3/latest/dev/acl-overview.html
        global_acl = 'http://acs.amazonaws.com/groups/global/AllUsers'

        def worker(bucket):
            try:
                bucket_result = {}
                bucket_acl = self.client.get_bucket_acl(Bucket=bucket['Name'])

                for grant in bucket_acl['Grants']:
                    if grant['Grantee']['Type'].lower() == 'group' \
                        and grant['Grantee']['URI'] == global_acl:
                        
                        if not bucket_result: bucket_result = bucket
                        bucket_result['Permissions'] = []
                        bucket_result['Permissions'].append(grant['Permission'])
                
                if bucket_result: results.append(bucket_result)

            # AccessDenied getting GetBucketAcl
            except ClientError: pass

        config = self.get_client_vars()

        for bucket in self.get_buckets(): self.pool.add_task(worker, bucket=bucket)
        self.pool.wait_completion()

        return self.inject_client_vars(results, config)

    def list_bucket_objects(self, bucket):
        '''List objects stored in a bucket
        
        Args:
            bucket (str): Name of the bucket
        
        Returns:
            list: List of bucket objects
        '''
        return self.client.list_objects(Bucket=bucket).get('Contents', [])
        



    def __init__(self):
        AwsBase.__init__(self, 's3')
